from ralium._util import TypeAlias, Literal, Any
from ralium.window import Window
from bs4 import BeautifulSoup

RaliumIDStr: TypeAlias = str
HTMLElementStr: TypeAlias = str

ALL_HTML_ELEMENTS: list[HTMLElementStr]
ALL_HTML_ELEMENT_INSTANCE_METHODS: list[str]
RALIUM_ID_IDENTIFIER_PREFIX: str

def create_ralium_id(tag: dict[str, Any]) -> str: ...
def create_element(
    html: BeautifulSoup, 
    tag: HTMLElementStr, 
    setitems: dict[str] | None = None, 
    **setattrs: dict[str, Any]
) -> None: ...

class JS:
    @staticmethod
    def str(__value: str) -> str: ...

    true: Literal['"true"']
    false: Literal['"false"']

    @staticmethod
    def bool(__value: str) -> bool | None: ...

    class ClassList(list):
        def __init__(self, element: HTMLElement) -> None: ...
        def add(self, classname: str) -> None: ...
        def remove(self, classname: str) -> None: ...

HTMLElementAttributes: list[str]

class HTMLElement:
    def __init__(self, window: Window, id: RaliumIDStr, element: HTMLElement = None) -> None: ...
    def __getattr__(self, identifier: str) -> Any: ...
    def __setattr__(self, identifier: str, value: Any) -> None: ...
    def _get_js_element(self) -> str: ...
    def _as_js_str(self) -> str: ...
    def _eval(self, cmd: str) -> Any: ...
    def _parse(self, value: str) -> Any: ...


from types import NoneType

class WindowConfig:
    """
    A Window Configuration Object.

    :param title: The Window Title.
    :param icon: The Window Icon.
    :param size: The Window Size.
    :param min_size: The minimum size the Window can be.
    :param resizable: If the window can be resized.
    :param builtins: Add builtin functions to the JavaScript function space.
    :param kwargs: Extra arguments for the `pywebview.create_window` function.

    :raises TypeError: If a specific config option has an invalid value type.
    """

    def __init__(self,
        title = None,
        icon = None,
        size = (900, 600), 
        min_size = (300, 300), 
        child_window = False, 
        resizable = True,
        builtins = True,
        **kwargs
    ):
        self.icon = icon
        self.size = size
        self.title = title
        self.min_size = min_size
        self.use_builtins = builtins
        self.is_resizable = resizable
        self.is_child_window = child_window
        self.other_options = kwargs

        if not isinstance(title, (str, NoneType)):
            raise TypeError(f"Expected the title to be a string, instead got '{type(title)}'")
        
        if not isinstance(size, tuple):
            raise TypeError(f"Expected the window size to be a tuple, instead got '{type(size)}'")
        
        if len(size) != 2:
            raise TypeError(f"Invalid size, expected just width and height.")
        
        if not isinstance(min_size, tuple):
            raise TypeError(f"Expected the minimum window size to be a tuple, instead got '{type(min_size)}'")
        
        if len(min_size) != 2:
            raise TypeError(f"Invalid minimum size, expected just width and height.")

        self.width, self.height = size
        self.min_width, self.min_height = min_size

        if not isinstance(self.width, int):
            raise TypeError(f"Expected the window width to be an int, instead got '{type(self.width)}'")
        
        if not isinstance(self.height, int):
            raise TypeError(f"Expected the window height to be an int, instead got '{type(self.height)}'")
        
        if not isinstance(self.min_width, int):
            raise TypeError(f"Expected the minimum window width to be an int, instead got '{type(self.min_width)}'")
        
        if not isinstance(self.min_height, int):
            raise TypeError(f"Expected the minimum window height to be an int, instead got '{type(self.min_height)}'")
        
        if not isinstance(self.is_child_window, bool):
            raise TypeError(f"Expected a boolean for argument 'child_window', instead got '{type(self.child_window)}'")
        
        if not isinstance(self.is_resizable, bool):
            raise TypeError(f"Expected a boolean for argument 'resizable', instead got '{type(self.child_window)}'")
    
    def as_webview_kwargs(self):
        """
        A helper function for converting the configuration options to a dict for `pywebview.create_window` function.
        """

        return {
            "title": self.title,
            "width": self.width,
            "height": self.height,
            "min_size": self.min_size,
            "resizable": self.is_resizable
        }


from .element import HTMLElement
from .webpage import WebHook, create_namespace
from .errors import DisableWarnings
from .config import WindowConfig
from .window import Window
from .setup import setup
from ._util import __version__

from . import errors
from . import api

from .builtins import (
    get_window_port,
    get_window_path,
    redirect,
    shutdown
)


from ralium.errors import *
from ralium._util import *
import inspect

HTML_TEMPLATE = """<!DOCTYPE html>
<html>
<head lang="en">
<meta charset="UTF-8">
</head>
<body>
</body>
</html>"""

class FileReader:
    def __init__(self, file, encoding):
        self.__data = file
        self.__read = False
        self.__encoding = encoding

        path_length = len(self.__data)
        path_limit = _get_path_limit()

        if path_length > path_limit:
            raise FilePathLimitError(f"Failed to load file, exceeds the max path limit of '{path_limit}'")
        
    @property
    def content(self):
        self._read()
        return self.__data
    
    def _read(self):
        if self.__read:
            return

        if not _check_exists(self.__data):
            return warn(f"Failed to load file with path: '{self.__data}'", FileNotFoundWarning, True)
        
        self.__data = _read_file(_get_path(self.__data), self.__encoding)
        self.__read = True

class CSSReader:
    def __init__(self, *files, encoding):
        self.__readers = []

        for file in files:
            if _check_exists(file):
                if len(file) <= _get_path_limit():
                    warn(f"File '{file}' exceeds path limit.", FilePathLimitWarning, major=True)
                
                self.__readers.append(FileReader(file, encoding))
                continue

            self.__readers.append(file)

    @property
    def content(self):
        items = []
        for reader in self.__readers:
            if isinstance(reader, FileReader):
                items.append(reader.content)
                continue
            items.append(reader)
        return "\n".join(items)

class WebHookNamespace(dict):
    def __init__(self, name, *functions, **named_functions):
        self.name = name

        self.add_functions(*functions)
        self.add_named_functions(**named_functions)
    
    def add_functions(self, *functions):
        for function in functions:
            if getattr(function, "_module_api_class", False):
                self.add_functions(*function.functions)
                self.add_named_functions(**function.named_functions)
                continue
            
            if not inspect.isfunction(function) and not inspect.ismethod(function):
                raise TypeError(f"Expected a function for namespace, instead got '{type(function)}'")

            self.__setitem__(function.__name__, function)
    
    def add_named_functions(self, **functions):
        for funcname, function in functions.items():
            if not inspect.isfunction(function) and not inspect.ismethod(function):
                raise TypeError(f"Expected a function for namespace, instead got '{type(function)}'")

            self.__setitem__(funcname, function)
    
class WebHookFunction:
    def __new__(cls, function, window):
        def wrapper(*args, **kwargs):
            return function(window, *args, **kwargs)
        
        wrapper.__name__     = function.__name__
        wrapper.__qualname__ = function.__name__

        return wrapper

class WebHook:
    """
    A WebHook Object that contains information for handling a specific URL.

    :param url: The URL that the WebHook handles.
    :param html: A file or raw text of the HTML to render.
    :param css: A file path or list of file paths to style the HTML.
    :param functions: Functions to expose to JavaScript.
    :param namespaces: Namespaces to expose to JavaScript.
    :param homepage: If this WebHook is a homepage, it will be the fallback page if something goes wrong with the `Navigation` handler.
    :param encoding: The file encoding of the HTML and CSS files.

    :raises FileNotFoundWarning: Displays a warning if a file doesn't exist. (Only if warnings are enabled.)
    :raises FilePathLimitError: If a path exceeds the system limit for path lengths.
    """

    def __init__(self, url, html, css = None, functions = None, namespaces = None, homepage = False, encoding = "UTF-8"):
        self.url = _norm_url(url)
        self.css = css or ""
        self.html = html
        self.window = None
        self.elements = []
        self.homepage = homepage
        self.encoding = encoding
        self.functions = functions
        self.namespaces = namespaces

        self._get_html()
        self._get_css()
    
    def __repr__(self):
        return f"WebHook(url='{self.url}')"
    
    def _get_html(self):
        if not _check_exists(self.html): return
        self.html = FileReader(self.html, self.encoding).content
    
    def _get_css(self):
        if isinstance(self.css, (list, tuple, set,)):
            files = [item for item in self.css if _check_exists(item)]
            values = [item for item in self.css if not _check_exists(item)]
            self.css = "\n".join([CSSReader(*files, encoding=self.encoding).content, *values])
            return
        
        if isinstance(self.css, (str, bytes,)) and _check_exists(self.css):
            self.css = CSSReader(self.css, encoding=self.encoding).content
    
    def _wrap_functions(self):
        if self.window is None:
            return

        self.functions = [WebHookFunction(function, self.window) for function in self.functions]
    
    def _wrap_namespaces(self):
        if self.window is None:
            return
        
        for namespace in self.namespaces:
            for key in namespace.keys():
                namespace[key] = WebHookFunction(namespace[key], self.window)

    def set_window(self, window):
        """Sets the window of the WebHook."""
        self.window = window

class WebHookDict(dict):
    def __init__(self, webhooks):
        for webhook in webhooks:
            self[webhook.url] = webhook

    def __iter__(self):
        return self.items().__iter__

    def __repr__(self):
        return f"{{{', '.join([repr(webhook) for webhook in self.values()])}}}"
    
    def get(self, url):
        webhook = super().get(url, None)

        if webhook is None:
            raise WebHookNotFoundError(f"Failed to find WebHook for the url '{url}'")

        return webhook

def create_namespace(name, *functions, **named_functions):
    """
    Creates a `Namespace` to use within JavaScript.

    :param name: The alias the namespace is accessed by.
    :param functions: A list of functions to add.
    :param named_functions: A dictionary of functions to add.

    :returns: A `WebHookNamespace` object.

    The names of the regularly listed functions will be the value
    of the `__name__` property of the function objects.

    The names of the named function dictionary will be the keyword part.
    """

    return WebHookNamespace(name, *functions, **named_functions)


from ralium.element import HTMLElement
from ralium.window import Window
from ralium._util import FunctionType, Callable, _RT

__registry__: list[FunctionType]

def _builtin(function: FunctionType) -> Callable[..., _RT]: ...

@_builtin
def redirect(window: Window, url: str) -> None: ...
@_builtin
def shutdown(window: Window) -> None: ...
@_builtin
def get_window_port(window: Window, element: HTMLElement | None = None) -> int: ...
@_builtin
def get_window_path(window: Window, element: HTMLElement | None = None) -> int: ...


from ralium.navigation import WindowNavigation
from ralium.listener import ClassListener
from ralium.webpage import WebHookDict, HTML_TEMPLATE
from ralium.config import WindowConfig
from ralium.engine import WebEngine
import ralium.builtins

from ralium.errors import (
    WebHookHomepageError, 
    WindowNotRunningError
)

from ralium.element import (
    HTMLElement,
    create_element, 
    create_ralium_id, 
    RALIUM_ID_IDENTIFIER_PREFIX
)

from bs4 import BeautifulSoup

import webview

@ClassListener
class Window:
    """
    Represents a Ralium GUI Window.

    :param webhooks: A list of WebHooks that handle each URL.
    :param config: Optional configuration for the window.

    :raises WebHookHompageError: If more than one WebHook has the homepage option set to `True`.
    """

    def __init__(self, webhooks, config = None):
        self.data = {}
        self.config = config or WindowConfig()
        self.engine = WebEngine(self)
        self.running = False
        self.webhooks = WebHookDict(webhooks)
        self.navigation = WindowNavigation()

        self.webview = webview.create_window(
            html=HTML_TEMPLATE,
            hidden=True,
            js_api=self.engine.create_api(),
            **self.config.as_webview_kwargs(),
            **self.config.other_options
        )

        for webhook in webhooks:
            if webhook.homepage:
                if self.navigation.homepage is not None:
                    raise WebHookHomepageError(f"A WebHook homepage already exists for the url '{self.navigation.homepage}'")
                self.navigation.setdefault(webhook.url)
            
            if self.config.use_builtins:
                webhook.functions.extend(ralium.builtins.__registry__)

            webhook.set_window(self)
            webhook._wrap_functions()
            webhook._wrap_namespaces()
        
        self.engine.start()
    
    def getElementsByClassName(self, className):
        """
        Gets all elements with a specific Class Name.

        :param className: The HTML Class to look for.

        :raises WindowNotRunningError: If you call this method before starting it, an error will be raised.

        :returns: A list of `HTMLElement` Objects. Returns an empty list if no elements are found.
        """

        if not self.running:
            raise WindowNotRunningError("Window must be running before calling the 'getElementsByClassName' method.")

        element_classes = self.webview.evaluate_js(f"Array.from(document.getElementsByClassName('{className}')).map((e) => e.classList);")
        return [HTMLElement(self, classList[0]) for classList in element_classes]
    
    def getElementById(self, id):
        """
        Gets an element with a specific Id.

        :param id: The HTML element Id.
        
        :returns: `HTMLElement` Object if found or `None`.
        """

        if self.running:
            element_ralium_id = str(self.webview.evaluate_js(f""" '' + document.getElementById("{id}").classList;"""))
            
            classes = element_ralium_id.split(' ')
            ralium_id = classes[0]

            if ralium_id not in RALIUM_ID_IDENTIFIER_PREFIX:
                for classname in classes:
                    if RALIUM_ID_IDENTIFIER_PREFIX not in classname:
                       continue

                    ralium_id = classname
                    break

            if ralium_id == "null":
                return

            return HTMLElement(self, ralium_id)

        soup = BeautifulSoup(self.webview.html, features="lxml")

        element = soup.select(f'#{id}')
        
        if not element:
            return
        
        ralium_id = element[0].get('class')[0]
        return HTMLElement(self, ralium_id)
    
    def load_handler(self, window):
        def on_load():
            window.show()
            window.events.loaded -= on_load

        window.events.loaded += on_load
        
    def display(self, url):
        """
        Displays a URL by calling the WebHook attached to the URL.

        :param url: The URL to display.
        """

        self.navigation.location = url
        webhook = self.webhooks.get(self.navigation.location)

        self.engine.bridge.clear()
        self.engine.functions.clear()

        for function in webhook.functions:
            self.engine.bridge.new(function)
            self.engine.functions.add_function(function)
        
        for namespace in webhook.namespaces:
            self.engine.bridge.new(namespace)

            for function in namespace.values():
                self.engine.functions.add_function(function)
        
        html = BeautifulSoup(webhook.html, features="lxml")
        webhook.elements = [create_ralium_id(element) for element in html.body.find_all()]

        create_element(html, "base", {"href": str(self.engine)})
        create_element(html, "script", string=str(self.engine.bridge))
        create_element(html, "style", string=webhook.css)

        self.render(html)

    def render(self, html):
        """Write HTML to the DOM."""
        if not self.running:
            self.webview.html = str(html)
            return

        self.webview.evaluate_js(f"document.open();document.write(String.raw`{str(html)}`);document.close();")
    
    def start(self):
        """Starts the Ralium Window and officially displays the GUI."""
        if self.running: return
        
        self.running = True
        webview.start(self.load_handler, self.webview, private_mode=False)
    
    def show(self):
        """Show a Ralium Window."""
        self.webview.show()
    
    def hide(self):
        """Hide a Ralium Window."""
        self.webview.hide()
    
    def shutdown(self):
        """Shutdown a Ralium Window."""
        self.webview.destroy()
        self.engine.close()


from ralium.bundle import BundledHTTPServer, FileSystem

from typing import (
    TypeAlias, 
    Callable, 
    TypeVar, 
    Literal,
    Self, 
    Any, 
)

from types import (
    FunctionType, 
    ModuleType
)

import http.server

__all__ = [
    "__version__",
    "BasicHTTPServer", "_get_http_server_handler", "_get_bundle", "_norm_url", "_check_exists", 
    "_check_is_dir", "_get_path_limit", "_read_file", "_get_path", "_is_markup_filelike"
]

__version__: str

_RT = TypeVar("_RT") # Return Type
ClassType: TypeAlias = object
DirPathStr: TypeAlias = str
FilePathStr: TypeAlias = str

class BasicHTTPServer(http.server.SimpleHTTPRequestHandler): ...

def _get_bundle() -> FileSystem | None: ...
def _get_http_server_handler() -> BasicHTTPServer | BundledHTTPServer: ...
def _check_exists(path: str) -> bool: ...
def _check_is_dir(path: str) -> bool: ...
def _norm_url(path: str) -> str: ...
def _get_path_limit_winreg() -> int: ...
def _get_path_limit() -> int: ...
def _read_file(path: str, encoding: str = "UTF-8") -> str: ...
def _get_path(path: str) -> str: ...
def _is_markup_filelike(markup: str) -> bool: ...


from ralium._util import FunctionType, ClassType, Any

class FunctionCallInfo:
    def __init__(self, 
        cls: ClassType, 
        function: FunctionType, 
        result: Any | None = None, 
        *args: tuple[Any], 
        **kwargs: dict[str, Any]
    ) -> None: ...
    def __repr__(self) -> str: ...

class ClassListener:
    class BroadcastFunction:
        def __init__(self) -> None: ...
        def __call__(self, *args, **kwargs) -> ClassType: ...
        def subscribe(self, function: FunctionType, is_after: bool = False) -> None: ...
        def unsubscribe(self, function: FunctionType, is_after: bool = False) -> None: ...

    def __init__(self, cls: ClassType) -> None: ...
    def __call__(self, *args: tuple[Any], **kwargs: dict[str, Any]) -> ClassType: ...
    
    def create_listener(self, function: FunctionType) -> BroadcastFunction: ...


from ralium._util import (
    FilePathStr,
    DirPathStr,
    Any
)

HTML_FILE_EXTENSIONS: list[str]
FILE_EXTENSIONS: list[str]

def _exe_str_error(name: str, var: Any) -> Exception: ...
def _add_data_arg(src: FilePathStr, dst: FilePathStr) -> str: ...

def collect_webfolder(webfolder: DirPathStr) -> dict[FilePathStr]: ...

def setup(
    pyfile: FilePathStr,
    name: str | None = None,
    icon: FilePathStr | None = None,
    bundle: bool = False,
    webfolder: DirPathStr | None = None,
    warnings: bool = False,
    onefile: bool = True,
    noconsole: bool = True,
    bundle_dist: DirPathStr | None = None,
    pyi_args: list[str] | None = None
) -> None: ...


from ralium.window import Window
from ralium._util import FunctionType
from typing import Iterator
import os

HTML_TEMPLATE: str

class FileReader:
    def __init__(self, file: str | os.PathLike, encoding: str) -> None: ...
    @property
    def content(self) -> str: ...
    def _read(self) -> None: ...

class CSSReader:
    def __init__(self, *files: list[str | os.PathLike], encoding: str) -> None: ...
    @property
    def content(self) -> str: None

class WebHookNamespace(dict):
    def __init__(self, 
        name: str, 
        *functions: tuple[FunctionType], 
        **named_functions: dict[str, FunctionType]
    ) -> None: ...

    def add_functions(self, *functions: tuple[FunctionType]) -> None: ...
    def add_named_functions(self, **functions: dict[str, FunctionType]) -> None: ...

class WebHookFunction:
    def __new__(cls, function: FunctionType, window: Window) -> function: ...

class WebHook:
    def __init__(self, 
        url: str, 
        html: str | os.PathLike, 
        css: str | list[str] | None = None, 
        functions: list[FunctionType] | None = None, 
        namespaces: list[WebHookNamespace] | None = None,
        homepage: bool = False, 
        encoding: str = "UTF-8"
    ) -> None: ...
    def __repr__(self) -> str: ...
    
    def _get_html(self) -> None: ...
    def _get_css(self) -> None: ...
    def _wrap_functions(self) -> None: ...
    def set_window(self, window: Window) -> None: ...

class WebHookDict(dict):
    def __init__(self, webhooks: list[WebHook]) -> None: ...
    def __repr__(self) -> str: ...
    def __iter__(self) -> Iterator[tuple]: ...
    def get(self, url: str) -> WebHook: ...

def create_namespace(
    name: str, 
    *functions: tuple[FunctionType], 
    **named_functions: dict[str, FunctionType]
) -> WebHookNamespace: ...


from typing import Any

class WindowConfig:
    def __init__(self,
        title: str | None = None,
        icon: str | None = None,
        size: tuple[int, int] = (900, 600), 
        min_size: tuple[int, int] = (300, 300), 
        child_window: bool = False, 
        resizable: bool = True,
        encoding: str = "UTF-8",
        builtins: bool = True,
        **kwargs
    ) -> None: ...

    def as_webview_kwargs(self) -> dict[str, Any]: ...


import uuid

ALL_HTML_ELEMENTS = [
    "a", "abbr", "acronym", "address", "area", "article", "aside", "audio", "b", "base",
    "bdi", "bdo", "big", "blockquote", "body", "br", "button", "canvas", "caption", "center",
    "cite", "code", "col", "colgroup", "data", "datalist", "dd", "del", "details", "dfn", "dialog",
    "dir", "div", "dl", "dt", "em", "embed", "fencedframe", "fieldset", "figcaption", "figure",
    "font", "footer", "form", "frame", "frameset", "h1", "head", "header", "hgroup", "hr", "html",
    "i", "iframe", "img", "input", "ins", "kbd", "label", "legend", "li", "link", "main", "map",
    "mark", "marquee", "menu", "menuitem", "meta", "meter", "nav", "nobr", "noembed", "noframes",
    "noscript", "object", "ol", "optgroup", "option", "output", "p", "param", "picture", "plaintext",
    "portal", "pre", "progress", "q", "rb", "rp", "rt", "rtc", "ruby", "s", "samp", "script", "search",
    "section", "select", "slot", "small", "source", "span", "strike", "strong", "style", "sub", "summary",
    "sup", "table", "tbody", "td", "template", "textarea", "tfoot", "th", "thead", "time", "title", "tr",
    "track", "tt", "u", "ul", "var", "video", "wbr", "xmp"
]

ALL_HTML_ELEMENT_INSTANCE_METHODS = [
    "after", "animate", "append", "attachShadow", "before", "checkVisibility", "closest",
    "computedStyleMap", "getAnimations", "getAttribute", "getAttributeNames", "getAttributeNode",
    "getAttributeNodeNS", "getAttributeNS", "getBoundingClientRect", "getClientRects",
    "getElementsByClassName", "getElementsByTagName", "getElementsByTagNameNS", "getHTML",
    "hasAttribute", "hasAttributeNS", "hasAttributes", "hasPointerCapture", "insertAdjacentElement",
    "insertAdjacentHTML", "insertAdjacentText", "matches", "prepend", "querySelector",
    "querySelectorAll", "releasePointerCapture", "remove", "removeAttribute", "removeAttributeNode",
    "removeAttributeNS", "replaceChildren", "replaceWith", "requestFullscreen", "requestPointerLock",
    "scroll", "scrollBy", "scrollIntoView", "scrollIntoViewIfNeeded", "scrollTo", "setAttribute",
    "setAttributeNode", "setAttributeNodeNS", "setAttributeNS", "setCapture", "setHTML",
    "setPointerCapture", "toggleAttribute"
]

RALIUM_ID_IDENTIFIER_PREFIX = "ralium-"

def create_ralium_id(tag):
    ralium_id = f"{RALIUM_ID_IDENTIFIER_PREFIX}{uuid.uuid1()}"

    tag['class'] = f"{ralium_id} {' '.join(tag.get('class', []))}".strip()

    return ralium_id

def create_element(html, tag, setitems = None, **setattrs):
    element = html.new_tag(tag)
    setitems = setitems or {}

    if not isinstance(setitems, dict):
        raise TypeError(f"Expected argument 'setitems' to be of type str, instead got '{type(setitems)}'")

    for name, value in setattrs.items():
        setattr(element, name, value)
    
    for name, value in setitems.items():
        element[name] = value
    
    html.body.append(element)

class JS:
    class _RawStr(str): pass

    @staticmethod
    def str(__value):
        return f'"{__value}"'
    
    @staticmethod
    def raw(__value):
        return JS._RawStr(__value)

    true = str("true")
    false = str("false")

    @staticmethod
    def bool(__value):
        if __value == "true":
            return True

        if __value == "false":
            return False

    class ClassList(list):
        def __init__(self, element):
            self.element = element
            super().__init__(str(element._eval(f"{element._as_js_str()}.classList")).split(' '))
        
        def add(self, classname):
            if classname in self:
                return

            super().append(classname)
            self.element._eval(f'{self.element._as_js_str()}.classList.add("{classname}")')
        
        def remove(self, classname):
            if classname not in self:
                return
            
            super().remove(classname)
            self.element._eval(f'{self.element._as_js_str()}.classList.remove("{classname}")')
    
HTMLElementAttributes = [
    "_ralium_id", 
    "_ralium_window", 
    "_ralium_element",
    "classList"
]

class HTMLElement:
    def __init__(self, window, id, element = None):
        self._ralium_id = id
        self._ralium_window = window
        self._ralium_element = element
        self.classList = JS.ClassList(self)

        if RALIUM_ID_IDENTIFIER_PREFIX in self._ralium_id or not id:
            return

        classes = str(self._eval(f""" '' + document.getElementById("{id}").classList;""")).split(' ')

        if RALIUM_ID_IDENTIFIER_PREFIX in classes[0]:
            self._ralium_id = classes[0]
            return

        for classname in classes:
            if RALIUM_ID_IDENTIFIER_PREFIX in classname:
                self._ralium_id = classname
        
    def __getattr__(self, identifier):
        if identifier in ALL_HTML_ELEMENT_INSTANCE_METHODS:
            def wrapper(*args, **kwargs):
                _prepare = lambda v: (str(v), f'"{v}"')[isinstance(v, str)]
                args_str = [_prepare(arg) for arg in args]
                args_str.extend([f'{_prepare(key)}={_prepare(value)}' for key, value in kwargs.items()])
                args_str = ",".join(args_str)

                return self._eval(f"{self._as_js_str()}.{identifier}({args_str})")
            
            wrapper.__name__     = identifier
            wrapper.__qualname__ = f"HTMLElement.{identifier}"

            return wrapper

        return self._parse(self._eval(f"{self._as_js_str()}.{identifier}"))
    
    def __setattr__(self, identifier, value):
        if identifier in HTMLElementAttributes:
            return object.__setattr__(self, identifier, value)
        
        if isinstance(value, str) and not isinstance(value, JS._RawStr):
            value = JS.str(value)

        self._eval(f"{self._get_js_element()}.{identifier} = {value};")
    
    def _get_js_element(self):
        return f"document.getElementsByClassName('{self._ralium_id}')[0]"
    
    def _as_js_str(self):
        return f"'' + {self._get_js_element()}"
    
    def _eval(self, cmd):
        return self._ralium_window.webview.evaluate_js(cmd)
    
    def _parse(self, value):
        if value in ["true", "false"]:
            return JS.bool(value)
        return value


from ralium.api import (
    _check_web_folder,
    _check_web_routes
)

from ralium._util import (
    SYS_BUNDLE_ATTRIBUTE,
    _get_bundle,
    _get_path,
    _norm_url
)

from shutil import COPY_BUFSIZE
from http import HTTPStatus

import urllib.parse
import http.server
import posixpath
import sys
import os

IMAGE_FILE_EXTENSIONS = [
    ".apng", ".gif", ".ico", ".cur", ".jpg", ".jpeg", 
    ".jfif", ".pjpeg", ".pjp", ".png", ".svg", ".ico",
    ".webp"
]

class BundledHTTPServer(http.server.SimpleHTTPRequestHandler):
    def do_GET(self):
        """Serve a GET request."""
        f = self.send_head()
        if f is not None:
            try:
                self.copyfile(f, self.wfile)
            except:
                return

    def do_HEAD(self):
        """Serve a HEAD request."""
        self.send_head()

    def send_head(self):
        """Common code for GET and HEAD commands.

        This sends the response code and MIME headers.

        Return value is either a file object (which has to be copied
        to the outputfile by the caller unless the command was HEAD,
        and must be closed by the caller under all circumstances), or
        None, in which case the caller has nothing further to do.

        """
        f = None
        path = self.translate_path(self.path)
        ctype = self.guess_type(path)
        # check for trailing "/" which should return 404. See Issue17324
        # The test for this was added in test_httpserver.py
        # However, some OS platforms accept a trailingSlash as a filename
        # See discussion on python-dev and Issue34711 regarding
        # parsing and rejection of filenames with a trailing slash
        if path.endswith("/"):
            self.send_error(HTTPStatus.NOT_FOUND, "File not found")
            return None
        try:
            f = _get_bundle()[path]
        except FileNotFoundError:
            self.send_error(HTTPStatus.NOT_FOUND, "File not found")
            return None

        try:
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-type", ctype)
            self.send_header("Content-Length", str(len(str(f))))
            self.end_headers()
            return f
        except:
            raise
    
    def translate_path(self, path):
        """Translate a /-separated PATH to the local filename syntax.

        Components that mean special things to the local file system
        (e.g. drive or directory names) are ignored.  (XXX They should
        probably be diagnosed.)

        """
        # abandon query parameters
        path = path.split('?',1)[0]
        path = path.split('#',1)[0]
        # Don't forget explicit trailing slash when normalizing. Issue17324
        trailing_slash = path.rstrip().endswith('/')
        try:
            path = urllib.parse.unquote(path, errors='surrogatepass')
        except UnicodeDecodeError:
            path = urllib.parse.unquote(path)
        path = posixpath.normpath(path)
        words = path.split('/')
        words = filter(None, words)
        path = "\\"
        for word in words:
            if os.path.dirname(word) or word in (os.curdir, os.pardir):
                # Ignore components that are not a simple file/directory name
                continue
            path = os.path.join(path, word)
        if trailing_slash:
            path += '/'
        return path

    def copyfile(self, fsrc, fdst):
        fdst_write = fdst.write
        while buf := fsrc[:COPY_BUFSIZE]:
            fdst_write(buf)

class File:
    def __init__(self, relpath, content):
        self.relpath = relpath
        self.content = content
    
    def __str__(self):
        return self.content.decode()

    def __repr__(self):
        return f"ralium.bundle.File(relpath='{self.relpath.replace('\\', '\\\\')}', content={repr(self.content)})"
    
class Bundle:
    def __init__(self, *, url, page, server, styles):
        self.url = url
        self.page = page
        self.server = server
        self.styles = styles
    
    def __repr__(self):
        return f"ralium.bundle.Bundle(url='{self.url}', page={repr(self.page)}, server={repr(self.server)}, styles={[repr(f) for f in self.styles]})"

class FileSystem(dict):
    def __init__(self, *, images, styles, bundles):
        self.images = images
        self.styles = styles
        self.bundles = bundles

        for image in images:
            self[image.relpath] = image.content
        
        for style in styles:
            self[style.relpath] = style.content
        
        for bundle in bundles:
            self[bundle.page.relpath] = bundle.page.content

            if bundle.server is not None:
                self[bundle.server.relpath] = bundle.server.content
            
            for style in bundle.styles:
                self[style.relpath] = style.content
    
    def __getitem__(self, __path):
        try:
            return super().__getitem__(__path)
        except KeyError:
            raise FileNotFoundError(f"Failed to find '{__path}'")

class PyBundler:
    def __init__(self, pyfile, webfolder):
        self.pyfile = pyfile
        self.webfolder = _get_path(webfolder)
        self.webroutes = os.path.normpath(os.path.join(webfolder, "routes"))
        self.cssfolder = os.path.normpath(os.path.join(webfolder, "styles"))
        self.imgfolder = os.path.normpath(os.path.join(webfolder, "images"))

        _check_web_folder(self.webfolder)
        _check_web_routes(self.webroutes)

    def view(self):
        images = self.collect(self.imgfolder, PyBundler.isimage)
        styles = self.collect(self.cssfolder, PyBundler.iscss)
        bundles = []

        for root, _, files in os.walk(self.webroutes):
            _url = _norm_url(root.split(self.webroutes)[-1])
            _page = None
            _server = None
            _styles = []

            for file in files:
                filepath = os.path.abspath(os.path.join(root, file))
                filename = os.path.basename(file)
                relpath = self.relpath(filepath)

                match filename:
                    case "+page.html":
                        _page = File(relpath, self.get_content(filepath))
                    case "+server.py":
                        _server = File(relpath, self.get_content(filepath))
                    case _:
                        _, ext = os.path.splitext(filename)
                        if ext != ".css": continue

                        with open(filepath, "rb") as f:
                            _styles.append(File(relpath, f.read()))
            
            bundles.append(Bundle(url=_url, page=_page, server=_server, styles=_styles))
        
        return [
            b"import ralium.bundle\n",
            b"import sys\n",
            f"setattr(sys, '{SYS_BUNDLE_ATTRIBUTE}', ralium.bundle.FileSystem(\n".encode(),
            f"    images = {images},\n".encode(),
            f"    styles = {styles},\n".encode(),
            f"    bundles = {bundles}\n".encode(),
            b"))\n\n",
            self.get_content(self.pyfile)
        ]
    
    def relpath(self, filename):
        return os.path.normpath(os.path.abspath(filename).removeprefix(self.webfolder))

    def collect(self, dir, callback):
        data = []

        for root, _, files in os.walk(dir):
            for file in files:
                filepath = os.path.abspath(os.path.join(root, file))
                if not callback(filepath): continue

                data.append(File(self.relpath(filepath), PyBundler.get_content(filepath)))
        
        return data

    @staticmethod
    def iscss(filename):
        _, ext = os.path.splitext(filename)
        return ext == ".css"
    
    @staticmethod
    def isimage(filename):
        _, ext = os.path.splitext(filename)
        return ext in IMAGE_FILE_EXTENSIONS
    
    @staticmethod
    def get_content(filename):
        if not os.path.exists(filename):
            return
        
        with open(filename, "rb") as f:
            return f.read()


from PyInstaller.__main__ import logger, run as CompileExe
from ralium.errors import DisableWarnings, SetupError
from ralium.bundle import PyBundler
from ralium._util import __version__
import sys
import os

HTML_FILE_EXTENSIONS = [".htm", ".html"]
FILE_EXTENSIONS = [*HTML_FILE_EXTENSIONS, ".css", ".py"]

def _exe_str_error(name, var):
    return TypeError(f"Expected executable {name} to be of type 'str', instead got '{type(var)}'")

def _add_data_arg(src, dst):
    return f'--add-data={src}{os.pathsep}{dst}'

def collect_webfolder(webfolder):
    """
    Collects all necessary files from a Ralium Web Folder used within the executable main Python file.

    :param webfolder: The path to the project folder.
    
    :returns: A dictionary where the keys contain the absolute paths and the values represent the relative paths.
    """

    webfiles = {}
    webfolder = os.path.abspath(webfolder)
    webroutes = os.path.join(webfolder, "routes")
    cssfolder = os.path.join(webfolder, "styles")
    imgfolder = os.path.join(webfolder, "images")
    webfolder_name = os.path.basename(webfolder)
    
    if os.path.exists(webroutes):
        for root, _, files in os.walk(webroutes):
            dest = os.path.normpath(f"{webfolder_name}\\{root.removeprefix(webfolder)}")
            webfiles.update({os.path.join(root, file): dest for file in files if os.path.splitext(file)[-1] in FILE_EXTENSIONS})
    
    if os.path.exists(cssfolder):
        for root, _, files in os.walk(cssfolder):
            dest = os.path.normpath(f"{webfolder_name}\\{root.removeprefix(webfolder)}")
            webfiles.update({os.path.join(root, file): dest for file in files if os.path.splitext(file)[-1] == ".css"})
    
    if os.path.exists(imgfolder):
        for root, _, files in os.walk(imgfolder):
            dest = os.path.normpath(f"{webfolder_name}\\{root.removeprefix(webfolder)}")
            webfiles.update({os.path.join(root, file): dest for file in files})
    
    return webfiles

def setup(
    pyfile,
    name = None,
    icon = None,
    bundle = False,
    webfolder = None,
    warnings = False,
    onefile = True,
    noconsole = True,
    bundle_dist = None,
    pyi_args = None
):
    """
    Compiles a Ralium project to an executable using PyInstaller.

    :param pyfile: Python file to compile.
    :param name: Display name for the executable.
    :param icon: Display icon for the executable.
    :param bundle: Bundles all of the html, css, python and image files into one executable. (Requires a `webfolder`)
    :param webfolder: Directory of the project.
    :param warnings: Calls the `DisableWarnings` to prevent warnings.
    :param onefile: Creates the executable as a standalone file.
    :param noconsole: Prevents a console from being displayed.
    :param bundle_dist: The directory name Ralium will use for bundling projects. (Default: `dist`)
    :param pyi_args: Extra parameters for PyInstaller to use.

    :raises TypeError: If the name or icon is not a `str` or `None`.
    :raises SetupError: If bundle is `True` while the webfolder is `None`.
    :raises RuntimeError: If this function is called within an already compiled executable file.
    :raises FileNotFoundError: If a certain file path doesn't exist.
    """

    logger.info("Ralium: %s", __version__)

    if bundle:
        if webfolder is None:
            raise SetupError("Cannot bundle project without a webfolder.")

        if bundle_dist is None:
            bundle_dist = "dist"
        
        bundle_dist = os.path.abspath(bundle_dist)

        if not os.path.exists(bundle_dist):
            os.mkdir(bundle_dist)
            logger.info("created %s", bundle_dist)

        name, ext = os.path.splitext(os.path.basename(pyfile))
        filename = os.path.join(bundle_dist, f"{name}.bundle{ext}")

        logger.info("Bundling '%s' with project '%s'", pyfile, os.path.abspath(webfolder))

        code = PyBundler(pyfile, webfolder).view()

        if not warnings:
            code.insert(0, b"import ralium.errors; ralium.errors.DisableWarnings()\n")

        with open(filename, "wb") as f:
            f.writelines(code)

        logger.info("wrote %s", filename)
        logger.info("Finished Bundling")

        pyfile = filename

    if getattr(sys, "frozen", False):
        raise RuntimeError("Ralium setup cannot be ran from an executable file.")
    
    if not os.path.exists(pyfile):
        raise FileNotFoundError(f"Failed to find python file '{pyfile}'")

    args = [pyfile, *(pyi_args or [])]

    if name is not None:
        if not isinstance(name, str):
            raise _exe_str_error("name", name)
        
        args.append(f"--name={name}")
    
    if icon is not None:
        if not isinstance(icon, str):
            raise _exe_str_error("icon", icon)

        if not os.path.exists(icon):
            raise FileNotFoundError(f"Failed to find icon file with path: '{icon}'")
        
        args.append(f"--icon={icon}")
    
    if onefile:
        args.append(f"--onefile")
    
    if not warnings:
        DisableWarnings()
    
    if noconsole:
        args.append(f"--noconsole")
    
    if webfolder and not bundle:
        for src, dst in collect_webfolder(webfolder).items():
            args.append(_add_data_arg(src, dst))

    CompileExe(args)


from ralium.webpage import WebHook, WebHookNamespace
from ralium.config import WindowConfig
from ralium.window import Window
from ralium.errors import *
from ralium._util import (
    _get_path, 
    _norm_url, 
    _read_file,
    _check_exists,
    _check_is_dir
)

from types import FunctionType, MethodType

import inspect
import sys
import os

RALIUM_API_REGISTRY_IDENTIFIER = "__ralium_registry__"
    
def _check_web_folder(webfolder):
    if not _check_exists(webfolder):
        raise WebFolderNotFoundError("Web folder path does not exist.")
    
    if not _check_is_dir(webfolder):
        raise WebFolderDirectoryError("Web folder must be a directory.")

def _check_web_routes(webroutes):
    if not _check_exists(webroutes):
        raise WebRoutesNotFoundError("Web folder must contain a routes directory.")
    
    if not _check_is_dir(webroutes):
        raise WebRoutesDirectoryError("Web folder routes must be a directory.")

def _collect_global_css_files(webfolder):
    cssfiles = []
    cssfolder = os.path.join(webfolder, "styles")

    if _check_exists(cssfolder) and _check_is_dir(cssfolder):
        for root, _, files in os.walk(webfolder):
            cssfiles.extend([os.path.join(root, file) for file in files if os.path.splitext(file)[-1] == ".css"])
    
    return cssfiles

def _get_registry_contents(pyfile):
    _globals = {}
    exec(_read_file(pyfile), _globals)

    return _globals.get(RALIUM_API_REGISTRY_IDENTIFIER, [])

def _bundle_collect_webhooks():
    filesystem = getattr(sys, "bundled")

    styles = [str(style) for style in filesystem.styles]
    webhooks = []

    for bundle in filesystem.bundles:
        _styles = [str(style) for style in bundle.styles]
        _styles.extend(styles)

        _functions = []
        _namespaces = []

        if bundle.server is not None:
            for item in _get_registry_contents(bundle.server.relpath):
                if isinstance(item, WebHookNamespace):
                    _namespaces.append(item)
                    continue
                
                if isinstance(item, (FunctionType, MethodType,)):
                    _functions.append(item)

        webhooks.append(WebHook(bundle.url, str(bundle.page), _styles, _functions, _namespaces, (bundle.url == "/")))
    
    return webhooks

def _collect_webhooks(webfolder, webroutes):
    if getattr(sys, "bundled", None) is not None:
        return _bundle_collect_webhooks()

    webhooks = []
    global_css = _collect_global_css_files(webfolder)
    root_split_string = os.path.join(os.path.basename(webfolder), "routes")

    for root, _, files in os.walk(webroutes):
        url = _norm_url(f"{root.split(root_split_string)[-1]}")

        if not files: 
            continue

        css = [*global_css]
        html = None
        functions = []
        namespaces = []

        for file in files:
            file_path = os.path.join(root, file)

            if file == "+page.html":
                html = file_path
                continue
                
            _, ext = os.path.splitext(file)
            if ext == ".css":
                css.append(file_path)
                continue

            if file == "+server.py":
                registry = _get_registry_contents(file_path)

                for item in registry:
                    if isinstance(item, WebHookNamespace):
                        namespaces.append(item)
                        continue
                    
                    if isinstance(item, (FunctionType, MethodType,)):
                        functions.append(item)

        webhooks.append(WebHook(url=url, html=html, css=css, homepage=(url == "/"), functions=functions, namespaces=namespaces))
    
    return webhooks

class Module:
    """
    A container, typically used for module methods, that wraps functions and named functions that don't need a window argument.

    :param functions: A list of functions or methods to wrap.
    :param named_functions: A dictionary of functions with names different from their current ones to wrap.
    """

    # Flag attribute used by the 'WebHookNamespace' class.
    _module_api_class = True
    
    def __init__(self, *functions, **named_functions):
        self.functions = [wrap(function) for function in functions]
        self.named_functions = {name: wrap(function) for name, function in named_functions.items()}

def create_window(webfolder, config = None):
    """
    Creates a window from a specific directory structure.

    :param webfolder: The project directory.
    :param config: The window configuration.

    :returns: A `Window` object.

    :raises WebFolderNotFoundError: If the webfolder path doesn't exist.
    :raises WebFolderDirectoryError: If the webfolder path is not a directory.
    :raises WebRoutesNotFoundError: If the `webfolder` directory doesn't contain a `routes` directory.
    :raises WebRoutesDirectoryError: If `{webfolder}/routes` is not a directory.
    """

    webfolder = _get_path(webfolder)
    webroutes = os.path.join(webfolder, "routes")

    _check_web_folder(webfolder)
    _check_web_routes(webroutes)

    webhooks = _collect_webhooks(webfolder, webroutes)

    return Window(webhooks, config)

def register():
    """
    Creates a registry list to expose functions and namespaces for use within JavaScript.

    :returns: A decorator for exposing Python functions to JavaScript.
    """

    globals = inspect.currentframe().f_back.f_globals
    globals[RALIUM_API_REGISTRY_IDENTIFIER] = []

    def wrapper(function):
        if not inspect.isfunction(function):
            raise WebFunctionApiError("WebFunction wrapper must be used on a function.")

        if function not in globals[RALIUM_API_REGISTRY_IDENTIFIER]:
            globals[RALIUM_API_REGISTRY_IDENTIFIER].append(function)
        
        def echo(*args, **kwargs):
            return function(*args, **kwargs)
        
        echo.__name__     = function.__name__
        echo.__qualname__ = function.__qualname__

        return echo

    return wrapper

def namespace(name, *functions, **named_functions):
    """
    Adds a namespace to the registry.

    :param name: An alias for referencing the namespace.
    :param functions: A list of functions or methods to register under the namespace.
    :param named_functions: A dictionary of functions with names different from their current ones to add to the namespace.
    """

    registry = inspect.currentframe().f_back.f_globals.get(RALIUM_API_REGISTRY_IDENTIFIER)

    if registry is None:
        raise RegistryNotFoundError(
            "No active registry found in the global space.",
            "Ensure that 'ralium.api.register' has been called to initialize the registry.",
            "('ralium.api.register' must be called at the top-level of the stack)"
        )
    
    registry.append(WebHookNamespace(name, *functions, **named_functions))

def wrap(function):
    """
    Wraps a function to accept a 'window' argument, typically for module methods.

    :param function: A function or method.

    :returns: A decorator for the wrapped function.
    """

    def wrapper(_, *args, **kwargs):
        return function(*args, **kwargs)
    
    wrapper.__name__     = function.__name__
    wrapper.__qualname__ = function.__qualname__
    
    return wrapper


from ralium.webpage import WebHookNamespace
from ralium.element import HTMLElement
from ralium.errors import BridgeEventError
from ralium import _util

from bs4 import BeautifulSoup

import socketserver
import threading
import inspect
import os

def _process_api_calldata(window, calldata):
    webhook = window.webhooks.get(window.navigation.location)
    new_calldata = []

    for data in calldata:
        if isinstance(data, dict):
            element_id = data.get("className", None)
            local_name = data.get("localName", None)

            if element_id is None or local_name is None or element_id not in webhook.elements: 
                continue

            soup = BeautifulSoup(window.webview.html, "html.parser")
            element = soup.find(local_name, {"class": element_id})

            new_calldata.append(HTMLElement(window, element_id, element))
            continue

        new_calldata.append(data)
    
    return new_calldata

class WebApiFunctionDict(dict):
    def add_function(self, function):
        self[str(function)] = function

class WebBridge:
    def __init__(self, *functions):
        self.bridges = []
        self.initial = "function bridge(func) {pywebview.api.bridge(func)};"

        for function in functions:
            self.new(function)
        
        self.clear = self.bridges.clear
    
    def __str__(self):
        return self.initial + "".join(self.bridges)
    
    def _function_template(self, name, string):
        return f"function {name}(...calldata){{return pywebview.api.bridge(\"{string}\", calldata)}}"

    def _namespace_template(self, alias, *functions, **named_functions):
        functions = [
            f'{function.__name__}: {self._function_template("", str(function))}'
            for function in functions
        ]

        functions.extend([
            f'{alias}: {self._function_template("", str(function))}'
            for alias, function in named_functions.items()
        ])

        return f"let {alias} = {{{','.join(functions)}}}"

    def new(self, obj):
        if inspect.isfunction(obj) or inspect.ismethod(obj):
            return self.bridges.append(self._function_template(obj.__name__, str(obj)))
        
        if isinstance(obj, WebHookNamespace):
            return self.bridges.append(self._namespace_template(obj.name, **dict(obj.items())))
    
class WebEngine:
    def __init__(self, window):
        self.port, worker = self._init_server()
        
        self.server = threading.Thread(target=worker, daemon=True)
        self.__running = threading.Event()

        self.event = self.create_event()
        self.bridge = WebBridge()
        self.functions = WebApiFunctionDict()

        # Strictly for the 'WebApi' class created by 'WebEngine.create_api'
        self._window = window

        self.start = self.server.start
    
    def __str__(self):
        return f"http://localhost:{self.port}/"

    @property
    def running(self):
        return not self.__running.is_set()

    def _init_server(self):
        httpd = socketserver.TCPServer(("", 0,), _util._get_http_server_handler())
        port = httpd.socket.getsockname()[1]

        def worker():
            while self.running:
                httpd.handle_request()
            os._exit(0)
        
        return port, worker
    
    def close(self):
        self.__running.set()
        self.server.join(timeout=5)
    
    def create_api(engine):
        class WebApi:
            def __init__(self):
                return

            def bridge(self, func, calldata=None):
                if not engine.functions.get(func, False): return

                calldata = _process_api_calldata(engine._window, calldata)

                if calldata:
                    return engine.functions[func](*calldata)
                
                return engine.functions[func]()
        
        return WebApi()
    
    def create_event(self):
        def BridgeEvent(function):
            if callable(function):
                if not str(function) in self.functions:
                    self.functions.add_function(function)
                return f"bridge('{str(function)}')"
            raise BridgeEventError("Event attribute is not a function.")
        return BridgeEvent


from ralium.element import JS

__registry__ = []

def _builtin(function):
    if function not in __registry__:
        __registry__.append(function)
    
    def wrapper(*args, **kwargs):
        return function(*args, **kwargs)
    
    return wrapper

@_builtin
def redirect(window, url):
    window.display(url)

@_builtin
def shutdown(window):
    window.shutdown()

@_builtin
def get_window_port(window, element = None):
    element.innerHTML = window.engine.port
    return window.engine.port

@_builtin
def get_window_path(window, element = None):
    element.innerHTML = JS.str(window.navigation.location)
    return window.navigation.location


from ralium._util import (
    FilePathStr,
    DirPathStr,
    Callable
)

from typing import BinaryIO
import http.server

IMAGE_FILE_EXTENSIONS: list[str]

class BundledHTTPServer(http.server.SimpleHTTPRequestHandler):
    def do_GET(self) -> None: ...
    def do_HEAD(self) -> None: ...
    def send_head(self) -> bytes | None: ...
    def translate_path(self, path) -> str: ...
    def copyfile(self, fsrc: bytes, fdst: BinaryIO) -> None: ...

class File:
    def __init__(self, relpath: str, content: bytes) -> None: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...

class Bundle:
    def __init__(self, *,
        url: str,
        page: bytes,
        server: bytes | None,
        styles: list[File]
    ) -> None: ...

    def __repr__(self) -> str: ...

class FileSystem(dict):
    def __init__(self, *,
        images: list[File],
        styles: list[File],
        bundles: list[Bundle]
    ) -> None: ...

    def __getitem__(self, __path: FilePathStr) -> bytes: ...

class PyBundler:
    def __init__(self, pyfile: FilePathStr, webfolder: DirPathStr) -> None: ...

    def view(self) -> list[bytes]: ...
    def relpath(self, filename: FilePathStr) -> FilePathStr: ...
    def collect(self, dir: DirPathStr, callback: Callable[[FilePathStr], bool]) -> list[File]: ...

    @staticmethod
    def iscss(filename: FilePathStr) -> bool: ...
    @staticmethod
    def isimage(filename: FilePathStr) -> bool: ...
    @staticmethod
    def get_content(filename: FilePathStr) -> bytes | None: ...